/***************************************************************************
                                 collector.cpp
                             -------------------
    begin                : Fri Feb 1 2002
    copyright            : (C) 2002 by Constantinos A. Kotsokalis
    email                : ckotso@grnet.gr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "collector.h"

Collector::Collector(){
	slotcount = lostpackets = 0;
	tsAdd = NULL;
	tsProc = NULL;
	tsPrev = NULL;
	tsLast = NULL;
	datapacket = new PDU();
	mtx = new Mutex();
}

Collector::~Collector(){
}


/** Uses the configuration directives stored in argument */
void Collector::setConfiguration(MConfig &config){
	// Get the configuration data from the Config object
	map< string, vector<string> > cmap = config.getconfigmap();
	
	for (map< string, vector<string> >::iterator it = cmap.begin(); it != cmap.end(); it++){
		// For every line get the key value and the associated vector with value arguments
		string key = (*it).first;
		vector<string> val = (*it).second;
		// If the vector is empty throw an error to stderr and exit
		if (val.empty()){
			cerr << "Key " << key << " is not associated with any value! (config file error?)" << endl;
			exit(0);
		}
		// Set the router from which we accept data
		if (key == "ROUTER"){
			const char *mstr = val[0].c_str();
			router = InetHostAddress(mstr);
			if (!router.isInetAddress()){
				cerr << "Router " << mstr << " is not a valid internet address" << endl;
				exit(0);
			}
		}
		// Set the host from which we accept web requests
		else if (key == "WEB_SERVER"){
			const char *mstr = val[0].c_str();
			websrv = InetHostAddress(mstr);
			if (!websrv.isInetAddress()){
				cerr << "Web server " << mstr << " is not a valid internet address" << endl;
				exit(0);
			}
		}
		// Set the timeslot duration
		else if (key == "TIMESLOT"){
			const char *mstr = val[0].c_str();
			SlotPeriod = atoi(mstr);
		}
		// Set the traffic increase threshold
		else if (key == "INCREASE"){
			const char *mstr = val[0].c_str();
			increase = atof(mstr);
		}
		// Set the traffic percentage threshold
		else if (key == "PERCENTAGE"){
			const char *mstr = val[0].c_str();
			percentage = atof(mstr);
		}
		// Set the person in charge of this detector
		else if (key == "OWNER"){
			incharge = (char *)val[0].c_str();
		}
		// Set the command to execute on attack start and stop
		else if (key == "EXECCMD"){
			execcmd = (char *)val[0].c_str();
		}
		// Set the port number on which we are listening for NF datagrams
		else if (key == "COLLECTOR_PORT"){
			const char *mstr = val[0].c_str();
			port = uint16_t(atoi(mstr));
		}
		// Set the port number on which we are listening for connections
		// from remote detectors
		else if (key == "PEER_PORT"){
			const char *mstr = val[0].c_str();
			peer_port = uint16_t(atoi(mstr));
		}
		// Set the port number on which we are listening for connections
		// from from the CGI script
		else if (key == "WEB_PORT"){
			const char *mstr = val[0].c_str();
			web_port = uint16_t(atoi(mstr));
		}
		// Set the command to execute when an attack is detected, to log it
		// in a database or other storage means -- NOT USED AT THE MOMENT
		else if (key == "DB_SCRIPT"){
			struct stat st;
			if (stat(val[0].c_str(), &st) == -1){
				perror(val[0].c_str());
				exit(0);
			}			
		}
		// Add a peer connection for a specific interface pair
		else if (key == "PEER_CONN"){
			const char *mstr = val[0].c_str();
			InetHostAddress local = InetHostAddress(mstr);
			if (!local.isInetAddress()){
				cerr << "Local endpoint " << mstr << " is not a valid internet address!" << endl;
				exit(0);
			}
			free((void *)mstr);
			int pos = val[1].find(':');
			string remdet, remport;
			if (pos != -1){
				remdet = string(val[1].substr(0, pos));
				remport = string(val[1].substr(pos+1));
			}
			else {
				cerr << "Error: Invalid port for " << val[1] << endl;
				exit(0);
			}
			mstr = remdet.c_str();
			const char *mstr2 = remport.c_str();
			setRemotePeer(local, InetHostAddress(mstr), atoi(mstr2));
		}
	}	
}

/** Does all the work (binds local socket, creates new threads, accepts data, etc) */
void Collector::BindAndLoop(){
	int count;
	uint16_t version, numflows;
	uint32_t curtime=0, unix_sec=0, unix_msec=0;
	uint32_t lastsec=0, lastmsec=0, tmplmsec = 0;
	uint8_t *tmpbuff = (uint8_t *)malloc(16);

	// Create and bind the local socket
	try {
		sock = UDPSocket(InetAddress("0"), port);
	}
	catch (Socket *socket) {
		perror("Can't bind UDP socket");
		exit(errno);
	}

	while (1) {
		// Check for data on the socket
		count = sock.Peek((void *)tmpbuff, 16);
		// Find out the version of the netflow datagram
		version = ntohs(*((uint16_t *)tmpbuff));
		// Find out the number of flows
		numflows = ntohs(*((uint16_t *)(tmpbuff+2)));
		// Find out sysuptime
		curtime = ntohl(*((uint32_t *)(tmpbuff+4)));
		// Find out current time (according to the router)
		unix_sec = ntohl(*((uint32_t *)(tmpbuff+8)));
		unix_msec = ntohl(*((uint32_t *)(tmpbuff+12))) / 1000 / 1000;

		switch (version){
			// If it's NetFlow version 1
			// Almost the same code with v5
			case NETFLOW_V1:
					if (sock.getPeer() != router){
						sock.Recv((void *)buff, V1_HEADER_LEN + numflows * V1_BODY_LEN);
						break;
					}
					if (!lastsec){
						makeSlots(unix_sec, unix_msec);
						lastsec = unix_sec + SlotPeriod;
						lastmsec = unix_msec;

						WebBroker *webbroker = new WebBroker(web_port, websrv, mtx, this);
						webbroker->Detach();
					}
					else if ((unix_sec > lastsec) || (unix_sec == lastsec && unix_msec > lastmsec) ){
						if (unix_sec >= lastsec + SlotPeriod)
							lastsec = unix_sec;
						tmplmsec = lastmsec + 1;
						makeSlots(lastsec, tmplmsec);
						lastsec = lastsec + SlotPeriod;
						slotcount++;
					}
					count = sock.Recv((void *)buff, V1_HEADER_LEN + numflows * V1_BODY_LEN);
					datapacket->dpSet(version, numflows, curtime, unix_sec, unix_msec, buff + V1_HEADER_LEN);
					datapacket->addFlows(tsAdd);
					break;
			// If it's NetFlow version 5
			case NETFLOW_V5:
					// If the router that sent the data is different from the router
					// from which we accept data, drop the datagram (receive it and break)
					if (sock.getPeer() != router){
						sock.Recv((void *)buff, V5_HEADER_LEN + numflows * V5_BODY_LEN);
						continue;
					}
					// If it's the first datagram we receive, create the new timeslots
					// for the interface pairs and advance the lasttime value by SlotPeriod
					if (!lastsec){
						makeSlots(unix_sec, unix_msec);
						lastsec = unix_sec + SlotPeriod;
						lastmsec = unix_msec;

						WebBroker *webbroker = new WebBroker(web_port, websrv, mtx, this);
						webbroker->Detach();
					}
					else if ((unix_sec > lastsec) || (unix_sec == lastsec && unix_msec > lastmsec) ){
						// If it's past SlotTime then create new timeslots for all pairs
						if (unix_sec >= lastsec + SlotPeriod)
							lastsec = unix_sec;
						tmplmsec = lastmsec + 1;
						makeSlots(lastsec, tmplmsec);
						// Advance the lasttime value
						lastsec = lastsec + SlotPeriod;
						slotcount++;
					}
					// Receive the datagram
					count = sock.Recv((void *)buff, V5_HEADER_LEN + numflows * V5_BODY_LEN);
					// Set the datapacket object content from the datagram we received
					datapacket->dpSet(version, numflows, curtime, unix_sec, unix_msec, buff + V5_HEADER_LEN);
					// Call the datapacket's addFlows() method to add flow statistics to the
					// dist hash table
					datapacket->addFlows(tsAdd);
					break;
			// If it's NetFlow version 8 (UNTESTED CODE!!!)
			// Almost the same code with v5
			case NETFLOW_V8:
					count = sock.Peek((void *)buff, V8_HEADER_LEN);
					uint8_t aggtype = *((uint8_t *)buff+22);
					uint16_t aggsize = 0;
					// Set the aggsize variable according to the aggregation method
					if (AS_ISSET(aggtype))
						aggsize = V8AS_BODY_LEN;
					else if (PP_ISSET(aggtype))
						aggsize = V8PP_BODY_LEN;
					else if (SP_ISSET(aggtype))
						aggsize = V8SP_BODY_LEN;
					else if (DP_ISSET(aggtype))
						aggsize = V8DP_BODY_LEN;
					else if (PR_ISSET(aggtype))
						aggsize = V8PR_BODY_LEN;
					if (sock.getPeer() != router){
						sock.Recv((void *)buff, V8_HEADER_LEN + numflows * aggsize);
						break;
					}
					if (!lastsec){
						makeSlots(unix_sec, unix_msec);
						lastsec = unix_sec + SlotPeriod;
						lastmsec = unix_msec;

						WebBroker *webbroker = new WebBroker(web_port, websrv, mtx, this);
						webbroker->Detach();
					}
					else if ((unix_sec > lastsec) || (unix_sec == lastsec && unix_msec > lastmsec) ){
						if (unix_sec >= lastsec + SlotPeriod)
							lastsec = unix_sec;
						tmplmsec = lastmsec + 1;
						makeSlots(lastsec, tmplmsec);
						lastsec = lastsec + SlotPeriod;
						slotcount++;
					}
					count = sock.Recv((void *)buff, V8_HEADER_LEN + numflows * aggsize);
					datapacket->dpSet(version, numflows, curtime, unix_sec, unix_msec, buff + V8_HEADER_LEN);
					datapacket->addFlows(tsAdd);
 					break;
		}
	}
}

/** Read property of char * ifinfofile. */
const char * Collector::getifinfofile(){
	return ifinfofile;
}

/** Write property of char * ifinfofile. */
void Collector::setifinfofile(const char * _newVal){
	ifinfofile = _newVal;
}

void Collector::setRemotePeer(const InetHostAddress& local, const InetHostAddress& remote, const uint16_t& remprt){
	RemoteDetector rmd;
	rmd.local = local;
	rmd.remote = remote;
	rmd.remport = remprt;

	RemotePeers.push_back(rmd);
}

RemoteDetector Collector::getRemotePeer(const InetHostAddress& local){
	for (vector<RemoteDetector>::iterator it = RemotePeers.begin(); it != RemotePeers.end(); it++)
		if ((*it).local == local){
			RemoteDetector rmd;
			rmd.local = (*it).local;
			rmd.remote = (*it).remote;
			rmd.remport = (*it).remport;
			return rmd;			
		}

	RemoteDetector rmd;
	rmd.remote = "0";
	rmd.remport = 0;
	return rmd;
}

void Collector::makeSlots(uint32_t& usec, uint32_t& umsec){
	mtx->EnterMutex();

	if (tsLast)
		delete tsLast;

	tsLast = tsProc;
	tsProc = tsPrev;

	if (tsLast){
		tsLast->setPrevious(NULL);
		tsLast->setPreviousAdd(NULL);
	}

	tsPrev = tsAdd;
	if (tsPrev)
		tsPrev->setPreviousAdd(NULL);

	TimeStamp tst(usec, umsec);	
	tsAdd = new TimeSlot(tst, SlotPeriod, getifinfofile());
	if (!tsAdd){
		perror("new");
		exit(errno);
	}
	tsAdd->setPrevious(tsPrev);
	tsAdd->setPreviousAdd(tsPrev);

	if (tsProc){
		tsProc->setPreviousAdd(NULL);
		if (tsProc->getPrevious()){
			Detector *detector = new Detector(tsProc, this, percentage, increase, string(incharge), string(execcmd), mtx);
			detector->Detach();
		}
	}

	mtx->LeaveMutex();
}


